/*
 * cat2html.c:
 * Converts a "cat" file with bold and underlined text achieved by overstriking
 * and turns it into HTML.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: cat2html.c,v 1.3 2002/11/11 01:07:22 chris Exp $";

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

struct charinfo {
    char c, f;
#define F_BOLD          1
#define F_UNDERLINE     2
#define F_STRIKEOUT     4
};

#define C_BACKSPACE     8       /* ^H */

char *html_entity(const char c) {
    static char buf[8];
    switch (c) {
        case '<':
            strcpy(buf, "&lt;");
            break;

        case '>':
            strcpy(buf, "&gt;");
            break;

        case '&':
            strcpy(buf, "&quot;");
            break;

        default:
            buf[0] = c;
            buf[1] = 0;
            break;
    }

    return buf;
}

int main(int argc, char **argv) {
    char *orig_text, *p;
    struct charinfo *chars, *q, *chars_end;
    size_t n_used = 0, n_alloc = 0;
    ssize_t n;
    int fd = 0;
    char *html_title = "cat2html output", f;

    if (argv[1])
        html_title = argv[1];

    orig_text = p = (char*)malloc(n_alloc = 16384);
    do {
        n = read(fd, p, n_alloc - (p - orig_text));
        if (n < 0 && errno != EINTR) {
            perror("cat2html: read");
            return 1;
        }
        
        p += n;
        n_used += n;
        if (p == orig_text + n_alloc) {
            orig_text = realloc(orig_text, n_alloc *= 2);
            p = orig_text + n_used;
        }
    } while (n);
    *p = 0;

    chars = calloc(n_used, sizeof(struct charinfo));

    /* Do markup. */
    for (p = orig_text, q = chars; p < orig_text + n_used; ++p) {
        if (*p == C_BACKSPACE && q > chars)
            --q;
        else {
            if (q->c) {
                if (strchr("_-", q->c)) {
                    char a = q->c;  /* Swap */
                    q->c = *p;
                    *p = a;
                }
                
                if (q->c == *p)
                    q->f |= F_BOLD;
                else if (*p == '_')
                    q->f |= F_UNDERLINE;
                else if (*p != ' ')
                    q->f |= F_STRIKEOUT;
                ++q;
            } else {
                q->c = *p;
                q->f = 0;
                ++q;
            }
        }
    }

    chars_end = q + 1;

    printf("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n"
           "<html>\n"
           "<!-- generated by cat2html, %s -->\n"
           "<head>\n"
           "<title>%s</title>\n"
           "</head>\n"
           "<body>\n"
           "<pre>\n",
           rcsid,
           html_title);
           
    
    /* Now emit the text as HTML. */
    for (q = chars, f = 0; q < chars_end; ++q) {
        if (f != q->f) {

            if ((f & F_BOLD)      && !(q->f & F_BOLD))
                printf("</strong>");
            if ((f & F_UNDERLINE) && !(q->f & F_UNDERLINE))
                printf("</em>");
            if ((f & F_STRIKEOUT) && !(q->f & F_STRIKEOUT))     /* XXX </strike> is deprecated. */
                printf("</strike>");

            if (!(f & F_BOLD)      && (q->f & F_BOLD))
                printf("<strong>");
            if (!(f & F_UNDERLINE) && (q->f & F_UNDERLINE))
                printf("<em>");
            if (!(f & F_STRIKEOUT) && (q->f & F_STRIKEOUT))     /* XXX <strike> is deprecated. */
                printf("<strike>");


            f = q->f;
        }
        printf("%s", html_entity(q->c));
    }

    printf("</pre>\n"
           "</body>\n"
           "</html>\n");

    return 0;
}
